"""Input output module."""

import json
from dataclasses import fields
from typing import Any, Dict
from uuid import UUID

from ambient.core import BaseElement
from ambient.simulation import Simulation


# don't use dataclasses asdict function since we don't want
# to recurse into each instance.
def _field_dict(obj: Any) -> Dict[str, Any]:
    return {
        "type": type(obj).__qualname__,
        "data": {
            f.name: obj.get_sensor(f.name) or getattr(obj, f.name) for f in fields(obj)
        },
    }


class SimulationEncoder(json.JSONEncoder):
    """Encode simulations as JSON."""

    def default(self, o: Any) -> Any:  # pylint: disable=method-hidden
        """Handle simulation specific classes."""
        if isinstance(o, UUID):
            return f"{o.urn}"

        if isinstance(o, Simulation):
            sim = _field_dict(o)
            sim["data"]["elements"] = {
                e.guid.urn: _field_dict(e) for e in o.elements.values()
            }
            return sim

        if isinstance(o, BaseElement):
            return o.guid

        return json.JSONEncoder.default(self, o)


def simulation_decoder(dct: Dict[str, Any]) -> Dict[str, Any]:
    """Rebuild simulation classes on import."""
    if "guid" in dct:
        dct["guid"] = UUID(dct["guid"])

    if "type" in dct and "data" in dct:
        factory = BaseElement.factory[dct["type"]]
        return factory(**dct["data"])

    return dct



"""Material instances for use."""

from ambient.material import Material


class Brickwork(Material):
    """Chen and Wang Table 1 Wall I."""

    conductivity = 0.840
    density = 1700
    specific_heat = 800


class Brickwork2(Material):
    """Chen and Wang Table 1 Wall III."""

    conductivity = 0.81
    density = 1800
    specific_heat = 880


class LimePlaster(Material):
    """Chen and Wang Table 1 Wall I."""

    conductivity = 0.7
    density = 1600
    specific_heat = 880


class HeavyweightConcrete(Material):
    """Chen and Wang Table 1 Wall III."""

    conductivity = 1.63
    density = 2300
    specific_heat = 1000


class Stucco(Material):
    """Chen and Wang Table 1 Wall II."""

    conductivity = 0.6924
    density = 1858
    specific_heat = 836.8  # note the values in the paper are incorrect


class HighDensityConcrete(Material):
    """Chen and Wang Table 1 Wall II."""

    conductivity = 1.731
    density = 2243
    specific_heat = 836.8  # note the values in the paper are incorrect


class Insulation(Material):
    """Chen and Wang Table 1 Wall II."""

    conductivity = 0.0433
    density = 32
    specific_heat = 836.8  # note the values in the paper are incorrect


class Plaster(Material):
    """Chen and Wang Table 1 Wall II."""

    conductivity = 0.7270
    density = 1602
    specific_heat = 836.8  # note the values in the paper are incorrect


class Aluminium(Material):
    """Generic aluminium properties."""

    conductivity = 30.0
    density = 3950
    specific_heat = 192.5



"""Material property classes."""

from abc import abstractmethod
from dataclasses import dataclass
from typing import ClassVar

import numpy as np

from ambient.core import BaseElement


@dataclass  # type: ignore
class MaterialBase(BaseElement):
    """Base material related information."""

    # XXX: subclasses must provide thermal_resistance attribute

    @abstractmethod
    def calculate_response_matrices(self, frequencies: np.ndarray) -> np.ndarray:
        """Calculate the frequency response for the material."""


@dataclass
class Material(MaterialBase):
    """Material related information."""

    conductivity: ClassVar[float] = np.inf  #: Material conductivity [W/m.K]
    density: ClassVar[float] = np.inf  #: Material density [kg/m^3]
    specific_heat: ClassVar[float] = np.inf  #: Material specific heat capacity [J/kg.K]
    thickness: float = np.inf  #: Material thickness [m]

    def __post_init__(self) -> None:
        """Set the thermal resistance of the material [m^2.K/W]."""
        self.thermal_resistance = self.thickness / self.conductivity

    def calculate_response_matrices(self, frequencies: np.ndarray) -> np.ndarray:
        """Calculate the frequency response for the material."""
        # common arg for sinh/cosh
        values = (
            self.thickness
            * np.sqrt(self.density * self.specific_heat / self.conductivity)
            * np.sqrt(frequencies * 1.0j),
        )

        # precompute sinh/cosh value
        sinh = np.sinh(values)
        cosh = np.cosh(values)

        # create an empty array of the correct size
        matrices = np.empty((frequencies.size, 2, 2), dtype=np.complex)

        # matrix components
        matrices[:, 0, 0] = cosh
        matrices[:, 0, 1] = self.thermal_resistance * sinh / values
        matrices[:, 1, 0] = values * sinh / self.thermal_resistance
        matrices[:, 1, 1] = cosh

        return matrices


@dataclass
class MaterialResistanceOnly(MaterialBase):
    """Resistance only material related information."""

    thermal_resistance: float = np.inf  #: Material resistance [m^2.K/W]

    def calculate_response_matrices(self, frequencies: np.ndarray) -> np.ndarray:
        """Calculate the frequency response for the material."""
        # create an empty array of the correct size
        matrices = np.ones((frequencies.size, 2, 2), dtype=np.complex)

        # matrix components
        matrices[:, 0, 1] = self.thermal_resistance
        matrices[:, 1, 0] = 0.0

        return matrices



"""Psychrometric functions."""

# This module is based on ASHRAE Fundamentals 2017. All references are to
# Chapter 1 unless noted otherwise.
# Units are as follows:
#   Temperature: Kelvin [K]
#   Pressure: Pascals [Pa]
#   Altitude: meters [m]
#   Humidity ratio: [kg_w / kg_da]
#   Specific enthalpy: [kJ / kg_da]
#   Specific volume: [m^3 / kg_da]
#   Mass flow rate: [kg / s]
#   Volume flow rate: [m3 / s]


from dataclasses import dataclass, field, InitVar
from enum import IntEnum

import numpy as np
import scipy.optimize

# refer to eq. 1
_DRY_AIR_GAS_CONSTANT = 8.314472 / 28.966

# refer to eq. 3
STANDARD_PRESSURE = 101325  #: Standard atmospheric pressure [Pa].

# refer to eq. 8
_MOLAR_MASS_RATIO = 18.015268 / 28.966

# make a special constant for saturated humidity ratio
SATURATED_HUMIDITY_RATIO = np.inf  #: Specifies humidity ratio at saturation.


class FlowType(IntEnum):
    """Type of air flow."""

    MASS = 1  #: Mass flow rate [kg / s].
    VOLUME = 2  #: Volumetric flow rate [m^3 / s].


class PropertyType(IntEnum):
    """Type of psychrometric property."""

    HUMIDITY_RATIO = 1  #: Humidity ratio [kg_w / kg_da].
    RELATIVE_HUMIDITY = 2  #: Relative humidity [%].
    SPECIFIC_ENTHALPY = 3  #: Specific enthalpy [kJ / kg_da].
    SPECIFIC_VOLUME = 4  #: Specific volume [m^3 / kg_da].
    WETBULB = 5  #: Wetbulb temperature [K].
    DEWPOINT = 6  #: Dewpoint temperature [K].


@dataclass
class MoistAir:
    """Class for psychrometric calculations."""

    # pylint: disable=too-many-instance-attributes

    property_type: InitVar[PropertyType]
    property_value: InitVar[np.ndarray]

    drybulb: np.ndarray
    humidity_ratio: np.ndarray = field(init=False, default=np.nan)

    flow_rate_type: InitVar[FlowType] = FlowType.MASS
    flow_rate_value: InitVar[np.ndarray] = None

    mass_flow_rate: np.ndarray = field(init=False, default=np.nan)

    pressure: float = STANDARD_PRESSURE

    def __post_init__(
        self,
        property_type: PropertyType,
        property_value: np.ndarray,
        flow_rate_type: FlowType,
        flow_rate_value: np.ndarray,
    ) -> None:
        """Initialise based on arguments."""
        if property_type is PropertyType.HUMIDITY_RATIO:
            # ensure this is an array
            self.humidity_ratio = np.asarray(property_value)
            self.drybulb = np.asarray(self.drybulb)

            # allow for special saturated state
            saturated_mask = self.humidity_ratio == SATURATED_HUMIDITY_RATIO
            if np.any(saturated_mask):
                self.humidity_ratio[saturated_mask] = humidity_ratio_saturation(
                    self.drybulb[saturated_mask], self.pressure
                )
        elif property_type is PropertyType.RELATIVE_HUMIDITY:
            # see eq. 23
            pressure_ratio = saturation_pressure(self.drybulb) / self.pressure
            degree_of_saturation = (
                property_value
                * (1.0 - pressure_ratio)
                / (1.0 - property_value * pressure_ratio)
            )
            self.humidity_ratio = degree_of_saturation * humidity_ratio_saturation(
                self.drybulb, self.pressure
            )
        elif property_type is PropertyType.SPECIFIC_ENTHALPY:
            # see eq. 30
            drybulb = kelvin_to_celsius(self.drybulb)
            self.humidity_ratio = (
                property_value - 1.006 * drybulb
            ) / enthalpy_water_vapour_saturated(self.drybulb)
        elif property_type is PropertyType.SPECIFIC_VOLUME:
            # see eq. 26
            self.humidity_ratio = (
                (self.pressure / 1000)
                * property_value
                / (_DRY_AIR_GAS_CONSTANT * self.drybulb)
                - 1.0
            ) / 1.607858
        elif property_type is PropertyType.WETBULB:
            # see eq. 31
            wetbulb_saturated_humidity_ratio = humidity_ratio_saturation(
                property_value, self.pressure
            )
            enth_water_sat = enthalpy_water_condensed_saturated(property_value)
            enth_wetbulb_sat = _specific_enthalpy_air(
                wetbulb_saturated_humidity_ratio, property_value
            )
            drybulb = kelvin_to_celsius(self.drybulb)
            self.humidity_ratio = (
                enth_wetbulb_sat
                - 1.006 * drybulb
                - wetbulb_saturated_humidity_ratio * enth_water_sat
            ) / (2501.0 - 1.86 * drybulb - enth_water_sat)
        elif property_type is PropertyType.DEWPOINT:
            # see eq. 36
            dewpoint_sat_pressure = saturation_pressure(property_value)
            self.humidity_ratio = (
                621.945
                * dewpoint_sat_pressure
                / (self.pressure - dewpoint_sat_pressure)
            )
        else:
            assert False, f"invalid property type: {property_type}"

        # set flow rates
        if flow_rate_type is FlowType.MASS:
            self.mass_flow_rate = flow_rate_value
        elif flow_rate_type is FlowType.VOLUME:
            self.mass_flow_rate = flow_rate_value / self.specific_volume

    def __mul__(self, other: "MoistAir") -> "MoistAir":
        """Multiply mass flow by float or array."""
        if isinstance(other, (float, np.ndarray)):
            return MoistAir(
                property_type=PropertyType.HUMIDITY_RATIO,
                property_value=self.humidity_ratio,
                drybulb=self.drybulb,
                flow_rate_type=FlowType.MASS,
                flow_rate_value=other * self.mass_flow_rate,
            )

        raise TypeError(f"must be float or numpy array not: {type(other)}")

    def __add__(self, other: "MoistAir") -> "MoistAir":
        """Add two instances based on mass and energy conservation."""
        if isinstance(other, MoistAir):
            assert self.mass_flow_rate is not None
            assert other.mass_flow_rate is not None
            assert (
                self.pressure == other.pressure
            ), f"pressures are not equal: {self.pressure} != {other.pressure}"

            mass_flow_total = self.mass_flow_rate + other.mass_flow_rate

            enthalpy_total = (
                self.mass_flow_rate * self.specific_enthalpy
                + other.mass_flow_rate * other.specific_enthalpy
            ) / mass_flow_total
            humidity_ratio_total = (
                self.mass_flow_rate * self.humidity_ratio
                + other.mass_flow_rate * other.humidity_ratio
            ) / mass_flow_total

            drybulb = celsius_to_kelvin(
                (enthalpy_total - 2501.0 * humidity_ratio_total)
                / (1.006 + 1.86 * humidity_ratio_total)
            )

            return MoistAir(
                property_type=PropertyType.HUMIDITY_RATIO,
                property_value=humidity_ratio_total,
                drybulb=drybulb,
                flow_rate_type=FlowType.MASS,
                flow_rate_value=mass_flow_total,
                pressure=self.pressure,
            )

        raise TypeError(f"must be MoistAir not: {type(other)}")

    @property
    def degree_of_saturation(self) -> float:
        """Calculate the degree of saturation of moist air."""
        return self.humidity_ratio / humidity_ratio_saturation(
            self.drybulb, self.pressure
        )

    @property
    def relative_humidity(self) -> float:
        """Calculate the relative humidity of moist air."""
        saturation = self.degree_of_saturation
        pressure_ws = saturation_pressure(self.drybulb)
        return saturation / (1.0 - (1.0 - saturation) * pressure_ws / self.pressure)

    @property
    def specific_volume(self) -> float:
        """Calculate the specific volume of moist air.

        References:
            2017 ASHRAE Handbook - Fundamentals, Chapter 1, Equation 26
        """
        return (
            _DRY_AIR_GAS_CONSTANT
            * self.drybulb
            * (1.0 + 1.607858 * self.humidity_ratio)
            / (self.pressure / 1000.0)
        )

    @property
    def specific_density(self) -> float:
        """Calculate the density of moist air."""
        return (1.0 + self.humidity_ratio) / self.specific_volume

    @property
    def specific_enthalpy(self) -> float:
        """Calculate the specific enthalpy of moist air."""
        return _specific_enthalpy_air(self.humidity_ratio, self.drybulb)

    @property
    def wetbulb(self) -> float:
        """Calculate the wet-bulb temperature of moist air."""

        def root_function(wetbulb: np.ndarray) -> np.ndarray:
            # refer to eq. 31
            humidity_ratio_wb_sat = humidity_ratio_saturation(wetbulb, self.pressure)
            return (
                self.specific_enthalpy
                + (humidity_ratio_wb_sat - self.humidity_ratio)
                * enthalpy_water_condensed_saturated(wetbulb)
                - _specific_enthalpy_air(humidity_ratio_wb_sat, wetbulb)
            )

        return scipy.optimize.root(root_function, self.drybulb).x

    @property
    def dewpoint(self) -> float:
        """Calculate the dew-point temperature of moist air."""
        # refer to eq. 36
        partial_pressure_water_dewpoint = partial_pressure_water(
            self.humidity_ratio, self.pressure
        )

        def root_function(dewpoint: np.ndarray) -> np.ndarray:
            return saturation_pressure(dewpoint) - partial_pressure_water_dewpoint

        return scipy.optimize.root(root_function, self.drybulb).x


def celsius_to_kelvin(temperature: np.ndarray) -> np.ndarray:
    """Convert celsius temperature to kelvin."""
    return temperature + 273.15


def kelvin_to_celsius(temperature: np.ndarray) -> np.ndarray:
    """Convert kelvin temperature to celsius."""
    return temperature - 273.15


def standard_pressure(altitude: np.ndarray) -> np.ndarray:
    """Calculate the standard pressure from altitude.

    Args:
        altitude: The altitude in meters.

    Returns:
        float: Standard pressure in Pa.
    """
    return 101325 * np.power(1.0 - 2.25577e-5 * altitude, 5.2559)


def standard_temperature(altitude: np.ndarray) -> np.ndarray:
    """Calculate the standard temperature from altitude.

    Args:
        altitude: The altitude in meters.

    Returns:
        float: Standard temperature in K.
    """
    return celsius_to_kelvin(15.0 - 0.0065 * altitude)


def _saturation_pressure_water(temperature: np.ndarray) -> np.ndarray:
    # Refer to RP-1485 eq. 2.91
    const_n = [
        np.nan,
        0.11670521452767e4,
        -0.72421316703206e6,
        -0.17073846940092e2,
        0.12020824702470e5,
        -0.32325550322333e7,
        0.14915108613530e2,
        -0.48232657361591e4,
        0.40511340542057e6,
        -0.23855557567849,
        0.65017534844798e3,
    ]
    theta = temperature + const_n[9] / (temperature - const_n[10])

    const_a = theta ** 2 + const_n[1] * theta + const_n[2]
    const_b = const_n[3] * theta ** 2 + const_n[4] * theta + const_n[5]
    const_c = const_n[6] * theta ** 2 + const_n[7] * theta + const_n[8]

    return (
        1.0e6
        * (2 * const_c / (-const_b + np.sqrt(const_b ** 2 - 4.0 * const_a * const_c)))
        ** 4.0
    )


def _saturation_pressure_ice(temperature: np.ndarray) -> np.ndarray:
    # Refer to RP-1485 eq. 2.92
    const_a = [np.nan, -0.212144006e2, 0.273203819e2, -0.610598130e1]
    const_b = [np.nan, 0.333333333e-2 - 1.0, 0.120666667e1 - 1.0, 0.170333333e1 - 1.0]

    theta = temperature / 273.16

    return 611.657 * np.exp(
        const_a[1] * theta ** const_b[1]
        + const_a[2] * theta ** const_b[2]
        + const_a[3] * theta ** const_b[3]
    )


def saturation_pressure(temperature: np.ndarray) -> np.ndarray:
    """Calculate the saturation pressure based on temperature.

    Args:
        temperature: Temperature in K.

    Returns:
        float: Saturation pressure in Pa.
    """
    temperature = np.asarray(temperature)
    pressure = np.empty_like(temperature)

    positive = temperature >= 0.0
    negative = temperature < 0.0

    pressure[negative] = _saturation_pressure_ice(temperature[negative])
    pressure[positive] = _saturation_pressure_water(temperature[positive])

    return pressure


def partial_pressure_water(
    humidity_ratio: np.ndarray, pressure: np.ndarray
) -> np.ndarray:
    """Calculate the partial pressure of water vapour."""
    return pressure * humidity_ratio / (_MOLAR_MASS_RATIO + humidity_ratio)


def humidity_ratio_saturation(
    temperature: np.ndarray, pressure: np.ndarray
) -> np.ndarray:
    """Calculate the saturation humidity ratio for a given temperature and pressure."""
    pressure_ws = saturation_pressure(temperature)
    return _MOLAR_MASS_RATIO * pressure_ws / (pressure - pressure_ws)


def _enthalpy_water_saturated_liquid(temperature: np.ndarray) -> np.ndarray:
    """Calculate the enthalpy of water as a saturated liquid.

    See also: eq. 32.
    """
    return 4.186 * kelvin_to_celsius(temperature)


def _enthalpy_water_saturated_solid(temperature: np.ndarray) -> np.ndarray:
    """Calculate the enthalpy of water as a solid (ice).

    See also: eq. 34.
    """
    return -333.4 + 2.1 * kelvin_to_celsius(temperature)


def enthalpy_water_condensed_saturated(temperature: np.ndarray) -> np.ndarray:
    """Calculate the enthalpy of water in saturated liquid or solid form.

    Args:
        temperature: Temperature of the water [K].

    Returns:
        float: Enthalpy of saturated water at the given temperature [kJ / kg_w].
    """
    temperature = np.asarray(temperature)
    enthalpy = np.empty_like(temperature)

    positive = temperature >= 0.0
    negative = temperature < 0.0

    enthalpy[negative] = _enthalpy_water_saturated_solid(temperature[negative])
    enthalpy[positive] = _enthalpy_water_saturated_liquid(temperature[positive])

    return enthalpy


def enthalpy_water_vapour_saturated(temperature: np.ndarray) -> np.ndarray:
    """Calculate the enthalpy of saturated water vapour.

    Args:
        temperature: Temperature of the water [K].

    Returns:
        float: Enthalpy of saturated water vapour at the given temperature [kJ / kg_w].
    """
    temperature = kelvin_to_celsius(temperature)
    return 2501 + 1.86 * temperature


def _specific_enthalpy_dry_air(drybulb: np.ndarray) -> np.ndarray:
    """Calculate the enthalpy from dry-bulb temperature."""
    drybulb = kelvin_to_celsius(drybulb)
    return 1.006 * drybulb


def _specific_enthalpy_air(
    humidity_ratio: np.ndarray, drybulb: np.ndarray
) -> np.ndarray:
    """Calculate the enthalpy from humidity ratio and dry-bulb temperature."""
    drybulb = kelvin_to_celsius(drybulb)
    return 1.006 * drybulb + humidity_ratio * (2501 + 1.86 * drybulb)



import numpy as np
import pytest

from ambient.construction import ConstructionLayered
from ambient.material import MaterialResistanceOnly
import ambient.material_data as mats


CTF_EPS = 0.02
HEAT_TRANSFER_EPS = 0.007
HEAT_TRANSFER_SUM_EPS = 1.0e-8

# fmt: off
SOL_AIR = 273.15 + np.array(
    (
        24.4, 24.4, 23.8, 23.3, 23.3, 23.8,
        25.5, 27.2, 29.4, 31.6, 33.8, 36.1,
        43.3, 49.4, 53.8, 55.0, 52.7, 45.5,
        30.5, 29.4, 28.3, 27.2, 26.1, 25.0,
    )
)
# fmt: on


def _calculate_sums(ctfs):
    """Calculate the ratio of sums based on eq. 41."""
    ctf_sum = np.sum(ctfs, axis=1)
    ctf_sum[:-1] /= ctf_sum[-1]

    return ctf_sum[:-1]


def iterate_heat_transfer(con, temp_in, temp_out, qi, qo=None, iterations=100):
    """Iterate the heat transfer calculation."""
    for iteration in range(iterations):
        for hour in range(24):
            qi[hour] = con.calculate_heat_flux_inside(temp_out, temp_in, qi, hour)
            if qo is not None:
                qo[hour] = con.calculate_heat_flux_outside(temp_out, temp_in, qo, hour)


def test_heat_transfer_wang_wall_iii():
    """Calculate properties of wall III from:

        Shengwei Wang, Youming Chen,
        Transient heat flow calculation for multilayer constructions using
            a frequency-domain regression method,
        Building and Environment,
        Volume 38, Issue 1,
        2003,
        Pages 45-61,
        ISSN 0360-1323,
        https://doi.org/10.1016/S0360-1323(02)00024-0.
    """
    con = ConstructionLayered(
        materials=[
            MaterialResistanceOnly(thermal_resistance=0.0586),
            mats.Stucco(thickness=25.39 / 1000),
            mats.HighDensityConcrete(thickness=101.59 / 1000),
            mats.Insulation(thickness=25.30 / 1000),
            mats.Plaster(thickness=19.05 / 1000),
            MaterialResistanceOnly(thermal_resistance=0.1206),
        ],
        timestep=3600,
    )

    # fmt: off
    qe_paper = np.array(
        (
            11.646,  9.999,  8.517,  7.203,  6.016,  4.960,
             4.081,  3.473,  3.209,  3.304,  3.755,  4.523,
             5.584,  7.162,  9.498, 12.441, 15.591, 18.406,
            20.260, 20.392, 19.019, 17.157, 15.247, 13.403,
        )
    )
    # fmt: on

    # initialise values
    qe = np.zeros(24)
    qo = np.zeros(24)
    temp_in = np.full_like(SOL_AIR, 273.15 + 24)
    temp_out = SOL_AIR

    # the paper values are after the 4th iteration
    iterate_heat_transfer(con, temp_in, temp_out, qe, qo, iterations=4)

    # check the heat transfer into the room
    assert np.allclose(qe, qe_paper, rtol=0.0, atol=0.02)
    assert np.allclose(qe, qe_paper, rtol=0.002, atol=0.0)

    # now do further iterations to check heat transfer convergence
    iterate_heat_transfer(con, temp_in, temp_out, qe, qo, iterations=96)

    # check inside and outside heat transfer
    assert np.allclose(
        np.sum(qe) + np.sum(qo), 0.0, rtol=0.0, atol=HEAT_TRANSFER_SUM_EPS
    )


def test_ctfs_wang_wall_iii():
    """Calculate properties of wall III from:

        Shengwei Wang, Youming Chen,
        Transient heat flow calculation for multilayer constructions using
            a frequency-domain regression method,
        Building and Environment,
        Volume 38, Issue 1,
        2003,
        Pages 45-61,
        ISSN 0360-1323,
        https://doi.org/10.1016/S0360-1323(02)00024-0.
    """
    con = ConstructionLayered(
        materials=[
            MaterialResistanceOnly(thermal_resistance=0.0586),
            mats.Stucco(thickness=25.39 / 1000),
            mats.HighDensityConcrete(thickness=101.59 / 1000),
            mats.Insulation(thickness=25.30 / 1000),
            mats.Plaster(thickness=19.05 / 1000),
            MaterialResistanceOnly(thermal_resistance=0.1206),
        ],
        timestep=3600,
    )

    # fix the order of the ctfs for comparison with paper values
    con._min_ctf_order = 5
    con._max_ctf_order = 5

    ctf_b = np.flip(np.array([0.002868, 0.053248, 0.060036, 0.007236, 0.000050, 0.0]))
    ctf_d = np.flip(np.array([1.0, -1.175710, 0.300608, -0.015605, 0.000005, 0.0]))

    # check the values of the bk and dk ctfs
    # note: only these values are check as we use the improved ctf method
    # so only the bk and dk values are expected to match
    assert np.allclose(np.round(con._ctfs[1], 6), ctf_b, rtol=CTF_EPS, atol=0.0)
    assert np.allclose(np.round(con._ctfs[3], 6), ctf_d, rtol=CTF_EPS, atol=0.0)

    # check all the ctf sums are close to K
    assert np.allclose(_calculate_sums(con._ctfs), con.thermal_transmittance)


def test_heat_transfer_wang_wall_iv():
    """Calculate properties of wall IV from:

        Shengwei Wang, Youming Chen,
        Transient heat flow calculation for multilayer constructions using
            a frequency-domain regression method,
        Building and Environment,
        Volume 38, Issue 1,
        2003,
        Pages 45-61,
        ISSN 0360-1323,
        https://doi.org/10.1016/S0360-1323(02)00024-0.
    """
    con = ConstructionLayered(
        materials=[
            MaterialResistanceOnly(thermal_resistance=0.060),
            mats.Brickwork(thickness=105 / 1000),
            MaterialResistanceOnly(thermal_resistance=0.180),
            mats.HeavyweightConcrete(thickness=100 / 1000),
            MaterialResistanceOnly(thermal_resistance=0.120),
        ],
        timestep=3600,
    )

    # fix the order of the ctfs for comparison with paper values
    con._min_ctf_order = 5
    con._max_ctf_order = 5

    # fmt: off
    qe_paper = np.array(
        (
            23.165, 21.194, 19.155, 17.146, 15.212, 13.375,
            11.671, 10.166,  8.955,  8.112,  7.690,  7.709,
             8.165,  9.107, 10.747, 13.210, 16.396, 19.982,
            23.461, 26.138, 27.378, 27.325, 26.407, 24.950,
        )
    )
    # fmt: on

    ctf_b = np.flip(
        np.array([0.000178, 0.013915, 0.043475, 0.018078, 0.001052, 0.000006])
    )
    ctf_d = np.flip(
        np.array([1.0, -1.619841, 0.724516, -0.064305, 0.001542, -0.000006])
    )

    # check the values of the bk and dk ctfs
    # note: only these values are check as we use the improved ctf method
    # so only the bk and dk values are expected to match
    assert np.allclose(np.round(con._ctfs[1], 6), ctf_b, rtol=CTF_EPS, atol=0.0)
    assert np.allclose(np.round(con._ctfs[3], 6), ctf_d, rtol=CTF_EPS, atol=0.0)

    # initialise values
    qe = np.zeros(24)
    qo = np.zeros(24)
    temp_in = np.full_like(SOL_AIR, 273.15 + 24)
    temp_out = SOL_AIR

    # the paper values are after the 4th iteration
    iterate_heat_transfer(con, temp_in, temp_out, qe, qo, iterations=4)

    # check the heat transfer into the room
    assert np.allclose(qe, qe_paper, rtol=0.0, atol=0.01)
    assert np.allclose(qe, qe_paper, rtol=0.002, atol=0.0)

    # now do further iterations to check heat transfer convergence
    iterate_heat_transfer(con, temp_in, temp_out, qe, qo, iterations=96)

    # check inside and outside heat transfer
    assert np.allclose(
        np.sum(qe) + np.sum(qo), 0.0, rtol=0.0, atol=HEAT_TRANSFER_SUM_EPS
    )

    # check all the ctf sums are close to K
    assert np.allclose(_calculate_sums(con._ctfs), con.thermal_transmittance)

    # check thermal transmittance value from paper
    assert np.around(con.thermal_transmittance, 5) == 1.83033


def test_construction_resistance_low():
    """Calculate properties of a low capacitance wall."""
    con = ConstructionLayered(
        materials=[
            MaterialResistanceOnly(thermal_resistance=0.060),
            mats.Aluminium(thickness=1 / 1000),
            MaterialResistanceOnly(thermal_resistance=0.020),
        ],
        timestep=3600,
    )

    # initialise values
    qe = np.zeros(24)
    qo = np.zeros(24)
    temp_in = np.full_like(SOL_AIR, 273.15 + 24)
    temp_out = SOL_AIR

    # use a large number of iterations for convergence
    iterate_heat_transfer(con, temp_in, temp_out, qe, qo, iterations=100)

    # check inside and outside heat transfer
    assert np.allclose(
        np.sum(qe) + np.sum(qo), 0.0, rtol=0.0, atol=HEAT_TRANSFER_SUM_EPS
    )

    # check all the ctf sums are close to K
    assert np.allclose(_calculate_sums(con._ctfs), con.thermal_transmittance)


def test_heat_transfer_chen_wall_i():
    """Wall I from Chen and Wang (2001) Appl. Math. Modelling 25."""
    con = ConstructionLayered(
        materials=[
            MaterialResistanceOnly(thermal_resistance=0.0546),
            mats.Brickwork2(thickness=240 / 1000),
            mats.LimePlaster(thickness=20 / 1000),
            MaterialResistanceOnly(thermal_resistance=0.1149),
        ],
        timestep=3600,
    )

    # fmt: off
    qe_paper = np.array(
        (
            26.2538, 24.2517, 22.1326, 20.0110, 17.9603, 16.0019,
            14.1543, 12.4624, 11.0116,  9.8995,  9.1978,  8.9494,
             9.1640,  9.8302, 11.0838, 13.1564, 16.1038, 19.7167,
            23.5651, 27.0298, 29.2700, 29.9109, 29.3332, 28.0139,
        )
    )
    # fmt: on

    # initialise values
    qe = np.zeros(24)
    qo = np.zeros(24)
    temp_in = np.full_like(SOL_AIR, 273.15 + 24)
    temp_out = SOL_AIR

    # the paper values are after the 4th iteration
    iterate_heat_transfer(con, temp_in, temp_out, qe, qo, iterations=4)

    # check the heat transfer into the room
    assert np.allclose(qe, qe_paper, rtol=HEAT_TRANSFER_EPS, atol=0.0)

    # now do further iterations to check heat transfer convergence
    iterate_heat_transfer(con, temp_in, temp_out, qe, qo, iterations=96)

    # check inside and outside heat transfer
    assert np.allclose(
        np.sum(qe) + np.sum(qo), 0.0, rtol=0.0, atol=HEAT_TRANSFER_SUM_EPS
    )

    # check all the ctf sums are close to K
    assert np.allclose(_calculate_sums(con._ctfs), con.thermal_transmittance)


def test_heat_transfer_resistance_only():
    """Wall I from Chen and Wang (2001) Appl. Math. Modelling 25."""
    con = ConstructionLayered(
        materials=[MaterialResistanceOnly(thermal_resistance=0.5)], timestep=3600
    )

    # initialise values
    qe = np.zeros(24)
    qo = np.zeros(24)
    temp_in = np.full_like(SOL_AIR, 273.15 + 24)
    temp_out = np.full_like(SOL_AIR, 273.15 + 26)

    # since there is no time history component, one iteration is sufficient
    iterate_heat_transfer(con, temp_in, temp_out, qe, qo, iterations=1)

    # check the heat transfer into the room
    assert np.allclose(qe, 4.0, rtol=0.0, atol=0.0)

    # now do further iterations to check heat transfer convergence
    iterate_heat_transfer(con, temp_in, temp_out, qe, qo, iterations=96)

    # check inside and outside heat transfer
    assert np.allclose(np.sum(qe) + np.sum(qo), 0.0, rtol=0.0, atol=0.0)

    # check all the ctf sums are close to K
    assert np.allclose(con.thermal_transmittance, 2.0)



"""Calculate solar data."""

from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Optional

import numpy as np

from ambient.core import BaseElement
from ambient.time import TimePeriodBase


@dataclass
class Location(BaseElement):
    """Location related information."""

    latitude: float = np.inf  #: location latitude [degrees]
    longitude: float = np.inf  #: location longitude [degrees]
    timezone: float = np.inf  #: time difference from UTC [hours]
    elevation: float = 0.0  #: elevation above sealevel [m]

    def __post_init__(self) -> None:
        """Do some checks on input."""
        assert self.latitude != np.inf
        assert self.longitude != np.inf
        assert self.timezone != np.inf


@dataclass  # type: ignore
class SolarBase(BaseElement):
    """Solar related calculation base class."""

    location: Optional[Location] = None  #: Location for the solar calculations
    timeperiod: Optional[TimePeriodBase] = None  #: Time period for solar calculations

    def __post_init__(self) -> None:
        """Do some checks on input."""
        assert self.location is not None
        assert self.timeperiod is not None

        self._extraterrestrial_solar = self._init_extraterrestrial_solar()
        self._equation_of_time = self._init_equation_of_time()
        self._declination = self._init_declination()
        self._apparent_solar_time = self._init_apparent_solar_time()
        self._hour_angle = self._init_hour_angle()
        self._solar_altitude = self._init_solar_altitude()
        self._solar_azimuth = self._init_solar_azimuth()

    def _init_extraterrestrial_solar(self) -> np.ndarray:
        assert self.timeperiod is not None
        # the extraterrestrial solar irradiation from eq. D.15
        # note: days are 1 based for calculations
        return 1367 * (
            1.0 + 0.033 * np.cos(2.0 * np.pi * (self.timeperiod.days - 2) / 365)
        )

    def _init_equation_of_time(self) -> np.ndarray:
        assert self.timeperiod is not None
        # note: days are 1 based for calculations
        gamma = 2 * np.pi * self.timeperiod.days / 365
        return 2.2918 * (
            0.0075
            + 0.1868 * np.cos(gamma)
            - 3.2077 * np.sin(gamma)
            - 1.4615 * np.cos(2 * gamma)
            - 4.089 * np.sin(2 * gamma)
        )

    def _init_declination(self) -> np.ndarray:
        assert self.timeperiod is not None
        # LCAM uses a simplified equation for declination
        # note: days are 1 based for calculations
        return np.radians(
            23.45 * np.sin(2 * np.pi * (self.timeperiod.days + 285) / 365)
        )

    def _init_apparent_solar_time(self) -> np.ndarray:
        assert self.timeperiod is not None
        assert self.location is not None
        standard_meridian = np.radians(self.location.timezone * 15)
        longitude = np.radians(self.location.longitude)

        return (
            self.timeperiod.times_in_hours[np.newaxis, :]
            + self._equation_of_time[:, np.newaxis] / 60.0
            + (standard_meridian - longitude) / 15.0
        )

    def _init_hour_angle(self) -> np.ndarray:
        return np.radians(15.0 * (self._apparent_solar_time - 12.0))

    def _init_solar_altitude(self) -> np.ndarray:
        assert self.location is not None
        latitude = np.radians(self.location.latitude)
        declination = self._declination

        solar_altitude = np.arcsin(
            np.cos(latitude)
            * np.cos(declination[:, np.newaxis])
            * np.cos(self._hour_angle)
            + np.sin(latitude) * np.sin(declination[:, np.newaxis])
        )
        solar_altitude[solar_altitude < 0.0] = 0.0

        return solar_altitude

    def _init_solar_azimuth(self) -> np.ndarray:
        assert self.location is not None
        latitude = np.radians(self.location.latitude)
        declination = self._declination
        hour_angle = self._hour_angle

        solar_azimuth = np.arccos(
            (
                np.sin(declination[:, np.newaxis]) * np.cos(latitude)
                - np.cos(declination[:, np.newaxis])
                * np.sin(latitude)
                * np.cos(self._hour_angle)
            )
            / np.cos(self._solar_altitude)
        )
        solar_azimuth[hour_angle > 0.0] *= -1.0
        solar_azimuth[hour_angle > 0.0] += 2.0 * np.pi

        return solar_azimuth

    @property
    @abstractmethod
    def diffuse_horizontal_irradiance_all(self) -> np.ndarray:
        """Return the diffuse horizontal irrandiance for all timesteps."""

    @property
    @abstractmethod
    def beam_normal_irradiance_all(self) -> np.ndarray:
        """Return the beam normal irrandiance for all timesteps."""

    def calculate_incident_diffuse_radiation(
        self, surface_azimuth: np.ndarray, surface_tilt: np.ndarray
    ) -> np.ndarray:
        """Calculate the incident diffuse radiation on surfaces.

        Args:
            surface_azimuth (float): surface azimuth values measured clockwise
                from North [degrees].
            surface_tilt (float): surface tilt values measured from horizontal [degrees]

        Returns:
            float: the incident diffuse solar radiation on each surface, for each day,
            for each hour [W / m^2]
        """
        surface_azimuth = surface_azimuth[np.newaxis, np.newaxis, :]
        surface_tilt = surface_tilt[np.newaxis, np.newaxis, :]
        solar_azimuth = self._solar_azimuth[..., np.newaxis]
        solar_altitude = self._solar_altitude[..., np.newaxis]

        # calculate the angle of incidence
        gamma = solar_azimuth - surface_azimuth
        cos_theta = np.cos(solar_altitude) * np.cos(gamma) * np.sin(
            surface_tilt
        ) + np.sin(solar_altitude) * np.cos(surface_tilt)

        # calculate the Y factor from eq. D.23
        y_factor = np.full_like(cos_theta, 0.45)
        y_factor[cos_theta >= -0.2] = 0.55 + 0.437 * cos_theta + 0.313 * cos_theta ** 2

        # calculate the incident diffuse radiant from eqs. D.24 and D.25
        incident_diffuse = y_factor * np.sin(surface_tilt)
        incident_diffuse[surface_tilt <= np.radians(90)] += np.cos(surface_tilt)
        incident_diffuse *= self.diffuse_horizontal_irradiance_all[..., np.newaxis]

        return incident_diffuse


@dataclass
class SolarDesignDay(SolarBase):
    """Solar related calculation design day."""

    tau_b: np.ndarray = field(
        default_factory=np.array
    )  #: clear sky beam optical depth on the 21st of each month []
    tau_d: np.ndarray = field(
        default_factory=np.array
    )  #: clear sky diffuse optical depth on the 21st of each month []

    def __post_init__(self) -> None:
        """Pre-compute data for annual calculations."""
        super().__post_init__()

        assert self.location is not None
        assert self.timeperiod is not None

        # calculate the beam solar
        m_const = 1.0 / (
            np.sin(self._solar_altitude)
            + 0.50572 * np.power((6.07995 + np.degrees(self._solar_altitude)), -1.6364)
        )
        m_const[self._solar_altitude < 0.0] = 0.0

        # calculate exponents
        ab_const = (
            1.454
            - 0.406 * self.tau_b
            - 0.268 * self.tau_d
            + 0.021 * self.tau_b * self.tau_d
        )
        ad_const = (
            0.507
            + 0.205 * self.tau_b
            - 0.080 * self.tau_d
            - 0.190 * self.tau_b * self.tau_d
        )

        # calculate the irradiance values
        self._beam_normal_irradiance = self._extraterrestrial_solar[
            :, np.newaxis
        ] * np.exp(
            -self.tau_b.take(self.timeperiod.months)[:, np.newaxis]
            * np.power(m_const, ab_const.take(self.timeperiod.months)[:, np.newaxis])
        )
        self._diffuse_horizontal_irradiance = self._extraterrestrial_solar[
            :, np.newaxis
        ] * np.exp(
            -self.tau_d.take(self.timeperiod.months)[:, np.newaxis]
            * np.power(m_const, ad_const.take(self.timeperiod.months)[:, np.newaxis])
        )

    @property
    def diffuse_horizontal_irradiance_all(self) -> np.ndarray:
        """Return the diffuse horizontal irrandiance for all timesteps."""
        return self._diffuse_horizontal_irradiance

    @property
    def beam_normal_irradiance_all(self) -> np.ndarray:
        """Return the beam normal irrandiance for all timesteps."""
        return self._beam_normal_irradiance



import json

import pytest

from ambient.core import BaseElement, SensorMISO
from ambient.construction import ConstructionLayered
from ambient.material_data import Brickwork, HeavyweightConcrete
from ambient.simulation import Simulation
from ambient.io import SimulationEncoder, simulation_decoder


def encode_simulation():
    """Test encoding a simulation."""
    sim = Simulation(name="Test sim")

    # register some materials
    mat1 = Brickwork(thickness=0.1)
    mat2 = HeavyweightConcrete(thickness=0.3)
    sensor = SensorMISO(source=[mat1, mat2])

    sim.register_elements([mat1, mat2, sensor])

    # register some layered constructions
    surfs = [ConstructionLayered(materials=sensor)]
    sim.register_elements(surfs)

    output = json.dumps(sim, cls=SimulationEncoder, indent=4, sort_keys=True)

    return sim, output


def test_decode_simulation():
    """Test decoding a simulation."""

    sim_input, output = encode_simulation()

    decoder = json.JSONDecoder(object_hook=simulation_decoder)
    sim = decoder.decode(output)
    sim.resolve_references()

    output = json.dumps(sim, cls=SimulationEncoder, indent=4, sort_keys=True)

    sim = decoder.decode(output)
    sim.resolve_references()

    assert sim_input == sim



"""Base classes."""

from abc import ABC, abstractmethod
from dataclasses import dataclass, field, fields
from uuid import UUID, uuid4
from typing import Any, ClassVar, Dict, List, Optional, Type

import numpy as np


_FUNCTION_MAP = {
    "AVG": np.average,
    "COUNT": np.size,
    "MAX": np.amax,
    "MIN": np.amin,
    "SUM": np.sum,
}


@dataclass
class BaseElement(ABC):
    """Implements required generic functionality for all elements."""

    factory: ClassVar[Dict[str, Type]] = {}

    guid: UUID = field(default_factory=uuid4)

    def __init_subclass__(cls: Type, **kwargs: Any) -> None:
        """Register the subclass for file loading."""
        super().__init_subclass__(**kwargs)  # type: ignore
        BaseElement.factory[cls.__qualname__] = cls

    def __getattribute__(self, name: str) -> None:
        """Allow sensors to be used as input to classes."""
        attr = super().__getattribute__(name)

        if isinstance(attr, SensorBase):
            return attr.value

        return attr

    def resolve_references(self, references: Dict[UUID, "BaseElement"] = None) -> None:
        """Resolve loaded guids within the simulation."""
        if references is None:
            return

        for fld in fields(self):
            if fld.name == "guid":
                continue

            attr = getattr(self, fld.name)

            if not isinstance(attr, str):
                continue

            try:
                attr = UUID(attr)
            except ValueError:
                continue

            setattr(self, fld.name, references[attr])

    def get_sensor(self, name: str) -> Any:
        """Retrieve the sensor connected to an attribute."""
        attr = super().__getattribute__(name)

        if isinstance(attr, SensorBase):
            return attr

        return None

    def get_dependencies(self) -> List["BaseElement"]:
        """Get all sensors which act as inputs to this element."""
        return [f for f in fields(self) if isinstance(f, SensorBase)]


@dataclass  # type: ignore
class SensorBase(BaseElement):
    """Base class for communication between elements."""

    @property
    @abstractmethod
    def value(self) -> Any:
        """Return the value of the sensor."""


@dataclass  # type: ignore
class SensorAttributeBase(SensorBase):
    """Sensor for retrieving a single attribute."""

    attribute: Optional[str] = None
    function: Optional[str] = None

    def __post_init__(self) -> None:
        """Check the input."""
        assert self.attribute or self.attribute is None
        assert self.function is None or self.function in _FUNCTION_MAP


@dataclass
class SensorSISO(SensorAttributeBase):
    """A sensor for extracting a single element attribute."""

    source: Optional[BaseElement] = None

    def __post_init__(self) -> None:
        """Check the input."""
        super().__post_init__()
        assert self.source
        assert isinstance(self.source, BaseElement)
        assert self.attribute is None or hasattr(self.source, self.attribute)

    def get_dependencies(self) -> List[BaseElement]:
        """All source elements are dependencies."""
        if self.source is None:
            return []

        if not isinstance(self.source, BaseElement):
            raise ValueError(
                "source is not a subclass of BaseElement. "
                "Have references been resolved?"
            )

        return [self.source]

    @property
    def value(self) -> Any:
        """Return the extracted value."""
        arg = self.source

        if self.attribute is not None:
            arg = getattr(arg, self.attribute)

        return arg if self.function is None else _FUNCTION_MAP[self.function](arg)


@dataclass
class SensorMISO(SensorAttributeBase):
    """A sensor for extracting and aggregating element attributes."""

    source: List[BaseElement] = field(default_factory=list)

    def __post_init__(self) -> None:
        """Check the input."""
        super().__post_init__()
        assert self.source
        assert isinstance(self.source, (list, tuple))
        assert self.attribute is None or all(
            hasattr(e, self.attribute) for e in self.source
        )

    def resolve_references(self, references: Dict[UUID, BaseElement] = None) -> None:
        """Resolve loaded guids within the simulation."""
        super().resolve_references(references)

        if references is None:
            return

        self.source = [
            element
            if isinstance(element, BaseElement)
            else references[UUID(str(element))]
            for element in self.source
        ]

    def get_dependencies(self) -> List[BaseElement]:
        """All source elements are dependencies."""
        if any(not isinstance(element, BaseElement) for element in self.source):
            raise ValueError(
                "an element of source is not a subclass of BaseElement. "
                "Have references been resolved?"
            )

        return self.source

    @property
    def value(self) -> Any:
        """Return the aggregated value."""
        args = self.source

        if self.attribute is not None:
            args = [getattr(e, self.attribute) for e in args]

        return args if self.function is None else _FUNCTION_MAP[self.function](args)



"""Simulation classes."""

from dataclasses import dataclass, field
from typing import Dict, Generator, List
from uuid import UUID

import networkx as nx

from ambient.core import BaseElement


@dataclass
class Simulation(BaseElement):
    """Base object holding all information for simulations."""

    name: str = ""
    elements: Dict[UUID, BaseElement] = field(default_factory=dict)

    def __post_init__(self) -> None:
        """Initialise other attributes."""
        self._dependency_graph = nx.DiGraph()

    def _register_element_no_check(self, element: BaseElement) -> None:
        """Include an element in the simulation without checking its type."""
        self.elements[element.guid] = element

    def register_elements(self, elements: List[BaseElement]) -> None:
        """Include a list of elements in a simulation."""
        if not all(isinstance(ele, BaseElement) for ele in elements):
            raise TypeError("all elements must be a subclass of BaseElement")

        for ele in elements:
            self._register_element_no_check(ele)

    def resolve_references(self, references: Dict[UUID, BaseElement] = None) -> None:
        """Resolve all element guids and recurse."""
        if references is not None:
            raise ValueError("references must be None")

        self.elements = {
            UUID(k) if isinstance(k, str) else k: v for k, v in self.elements.items()
        }

        for ele in self.elements.values():
            ele.resolve_references(self.elements)

    def create_dependency_graph(self) -> None:
        """Create the dependency graph."""
        for guid in self.elements:
            self._dependency_graph.add_node(guid)

        for ele in self.elements.values():
            deps = ele.get_dependencies()

            if deps is None:
                continue

            self._dependency_graph.add_edges_from((ele.guid, dep) for dep in deps)

    def evaluation_order(self) -> Generator:
        """Return the order for element evaluation."""
        return nx.topological_sort(self._dependency_graph)



"""Time based functionality."""

from calendar import isleap
from dataclasses import dataclass, field
from datetime import datetime
from enum import IntEnum
from typing import ClassVar, Type, TypeVar

import numpy as np

from ambient.core import BaseElement

HOUR_IN_SECONDS = 60 * 60
DAY_IN_HOURS = 24
DAY_IN_SECONDS = DAY_IN_HOURS * HOUR_IN_SECONDS

DEFAULT_TIMESTEPS_PER_HOUR = 1
DEFAULT_YEAR = 1900


T = TypeVar("T", bound="TimePeriodBase")  # pylint: disable=invalid-name


def _month_lengths(is_leap_year: bool = False) -> np.ndarray:
    month_lengths = np.array((31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31))

    if is_leap_year:
        month_lengths[1] = 29

    return month_lengths


def _month_lookup(month_lengths_array: np.ndarray) -> np.ndarray:
    return np.repeat(np.arange(12), month_lengths_array)


class DayType(IntEnum):
    """Enumerator for day types."""

    NORMAL = 1  #: Treat the days as normal days of the year.

    COOLING_DESIGN = 10  #: Treat the days as cooling design days.
    HEATING_DESIGN = 11  #: Treat the days as heating design days.
    VENTILATION_DESIGN = 12  #: Treat the days as ventilation design days.
    OTHER_DESIGN = 13  #: Treat the days as ventilation design days.


@dataclass
class TimePeriodBase(BaseElement):
    """Class for a simulation time period."""

    day_type: ClassVar[DayType] = DayType.NORMAL

    timesteps_per_hour: int = DEFAULT_TIMESTEPS_PER_HOUR
    days: np.ndarray = field(default_factory=np.array)
    year: int = DEFAULT_YEAR

    @classmethod
    def from_days_and_months(
        cls: Type[T],
        days: np.ndarray,
        months: np.ndarray,
        timesteps_per_hour: int = DEFAULT_TIMESTEPS_PER_HOUR,
        year: int = DEFAULT_YEAR,
    ) -> T:
        """Build a time period from days of the month and the months.

        Args:
            days (array_like): The days of the month.
            months (array_like): The months for each day.
            timesteps_per_hour (int): The number of timesteps per hour.
            year (int): The year for the days.

        Returns:
            TimePeriodBase: The time period with days derived from the input.
        """
        if isinstance(days, int):
            days = np.asarray([days])

        if isinstance(months, int):
            months = np.asarray([months])

        days, months = np.broadcast_arrays(days, months)
        offsets = np.cumsum(
            np.pad(_month_lengths(isleap(year)), (1, 0), constant_values=0)
        )
        days = np.take(offsets, months) + days

        return cls(timesteps_per_hour=timesteps_per_hour, days=days, year=year)

    def __post_init__(self) -> None:
        """Intialise other data."""
        assert self.days is not None

        self._timestep_length = HOUR_IN_SECONDS / self.timesteps_per_hour
        self._timestep_index = 0
        self._day_index = 0
        self._times = (
            np.arange(0, DAY_IN_HOURS * self.timesteps_per_hour)
            / self.timesteps_per_hour
        )
        self._months = np.fromiter(
            (
                datetime.strptime(f"{day + 1} {self.year}", "%j %Y").month - 1
                for day in self.days
            ),
            dtype=np.int,
        )

    @property
    def timestep(self) -> int:
        """Return the current timestep."""
        return self._timestep_index

    @property
    def times_in_hours(self) -> np.ndarray:
        """Return all timesteps for a day in hours."""
        return self._times

    @property
    def time_in_hours(self) -> float:
        """Return the current time of day in hours."""
        return self.times_in_hours[self._timestep_index]

    @property
    def day(self) -> int:
        """Return the current day of the year."""
        assert self.days is not None
        return self.days[self._day_index]

    @property
    def months(self) -> np.ndarray:
        """Return all the months of the year."""
        return self._months

    @property
    def month(self) -> int:
        """Return the current month of the year."""
        return self._months[self._day_index]

    @property
    def is_leap_year(self) -> bool:
        """Return true if the current year is a leap year."""
        return isleap(self.year)

    @property
    def total_timesteps(self) -> int:
        """Return the total number of timesteps from start to end."""
        assert self.days is not None
        return self.days.size * self.times_in_hours.size


@dataclass
class CoolingDesignDay(TimePeriodBase):
    """Cooling design day time period."""

    day_type: ClassVar[DayType] = DayType.COOLING_DESIGN


@dataclass
class HeatingDesignDay(TimePeriodBase):
    """Heating design day time period."""

    day_type: ClassVar[DayType] = DayType.HEATING_DESIGN



from datetime import datetime, timedelta

import numpy as np

from ambient.time import HeatingDesignDay


def test_total_timesteps():
    """Test the calculation of number of timesteps."""
    time = HeatingDesignDay(days=np.array((21,)))

    assert time.total_timesteps == 24


def test_total_day_month_constructor_jan():
    """Test the construction using days and months."""
    time = HeatingDesignDay.from_days_and_months(days=21, months=0)

    assert time.days == 21
    assert time.months == 0
    assert time.total_timesteps == 24


def test_total_day_month_constructor_feb():
    """Test the construction using days and months."""
    time = HeatingDesignDay.from_days_and_months(days=21, months=1)

    assert time.days == 52
    assert time.months == 1
    assert time.total_timesteps == 24



"""Constructions for simulation."""

# All references are to the paper:
# Transient heat flow calculation for multilayer constructions
# using a frequency-domain regression method
# Wang and Chen
# Building and Environment 38 (2003)

from abc import abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from functools import reduce
from typing import Dict, List, Tuple

import numpy as np

from numpy.polynomial.polynomial import polyzero, polyone

from ambient.core import BaseElement
from ambient.material import MaterialBase, MaterialResistanceOnly

_EPS = 1.0e-9


def _calculate_frequencies(
    lower_limit_power: int = -8, upper_limit_power: int = -3, num_freqs: int = 50
) -> np.ndarray:
    """Calculate the frequencies for the FDR method.

    References:
        Section 4, steps 1 and 2.
    """
    return np.logspace(lower_limit_power, upper_limit_power, num=num_freqs)


def _calculate_response_functions(response_matrix: np.ndarray) -> np.ndarray:
    """Calculate the response functions for FDR method.

    References:
        Section 2, equations 10, 11, and 12
    """
    response_functions = 1.0 / response_matrix[:, 0, 1]
    response_functions = np.tile(response_functions, (3, 1))
    response_functions[0] *= response_matrix[:, 0, 0]
    response_functions[2] *= response_matrix[:, 1, 1]

    return response_functions


def _approximate_response_function(
    response_function: np.ndarray,
    frequencies: np.ndarray,
    numerator_order: int,
    denominator_order: int,
    denominator_values: np.ndarray = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """Approximate the transfer functions by rational functions."""
    # calculate two parts of matrix from eq. 21
    max_power = max(numerator_order, denominator_order)
    powers = np.power(1.0j * frequencies[:, np.newaxis], np.arange(max_power + 1))

    h_matrix = np.copy(powers[:, : numerator_order + 1])

    g_vector = response_function

    # account for case of solving for denominator too
    if denominator_values is None:
        d_comp = -1.0 * np.copy(powers[:, 1 : denominator_order + 1])
        h_matrix = np.concatenate((h_matrix, d_comp * g_vector[:, np.newaxis]), axis=1)
    else:
        g_vector = g_vector * denominator_values

    gamma = np.real(np.conj(h_matrix.T) @ h_matrix)
    theta = 0.5 * np.real(
        np.conj(h_matrix.T) @ g_vector + h_matrix.T @ np.conj(g_vector)
    )

    coeffs = np.linalg.solve(gamma, theta)
    coeffs = np.insert(coeffs, numerator_order + 1, 1)

    return (
        np.flip(coeffs[: numerator_order + 1]),
        np.flip(coeffs[numerator_order + 1 :]),
    )


def _calculate_residuals(
    numerator_poly: np.poly, denominator_poly: np.poly, denominator_roots: np.ndarray
) -> np.ndarray:
    """See eq. 36."""
    delta = np.zeros_like(denominator_roots)
    for idx, root in enumerate(denominator_roots):
        root_poly = np.poly1d((1.0, -root))
        alpha_prime_coeff, rem = np.polydiv(denominator_poly, root_poly)

        assert np.allclose(rem, np.poly1d(polyzero), rtol=0.0, atol=1.0e-7)

        alpha_prime = np.poly1d(alpha_prime_coeff)

        delta_n = -1.0 * numerator_poly(root) / (root ** 2 * alpha_prime(root))
        delta[idx] = delta_n

    return delta


@dataclass  # type: ignore
class ConstructionBase(BaseElement):
    """Construction base class."""

    @property
    @abstractmethod
    def thermal_resistance(self) -> float:
        """Return the total thermal resistance of the construction."""

    @property
    def thermal_transmittance(self) -> float:
        """Return the total thermal transimittance of the construction."""
        return 1.0 / self.thermal_resistance

    @abstractmethod
    def calculate_heat_flux_inside(
        self,
        outside_temps: np.ndarray,
        inside_temps: np.ndarray,
        inside_heat_fluxes: np.ndarray,
        current_index: int,
    ) -> float:
        """Calculated the inside heat fluxes."""

    @abstractmethod
    def calculate_heat_flux_outside(
        self,
        outside_temps: np.ndarray,
        inside_temps: np.ndarray,
        outside_heat_fluxes: np.ndarray,
        current_index: int,
    ) -> float:
        """Calculated the inside heat fluxes."""


@dataclass
class ConstructionLayered(ConstructionBase):
    """Class for layered constructions.

    The order of layers is outside to inside.
    """

    materials: List[MaterialBase] = field(default_factory=list)
    timestep: int = 3600  #: The time step for simulation [s]

    def __post_init__(self) -> None:
        """Set construction data not stored in fields."""
        # allow fixing the minimum and maximum order of ctfs for testing
        self._min_ctf_order = 1
        self._max_ctf_order = 6
        self._ctfs_internal = None

    @property
    def thermal_resistance(self) -> float:
        """Return the thermal transimittance of the construction."""
        return sum(m.thermal_resistance or 0.0 for m in self.materials)

    @property
    def _is_resistance_only(self) -> bool:
        return all(isinstance(m, MaterialResistanceOnly) for m in self.materials)

    @property
    def _ctfs(self) -> np.ndarray:
        """Return the conduction transfer functions."""
        if self._ctfs_internal is None:
            # paper notes typical orders of 3-5, however 6 may be required for heavy
            # constructions. Additionally, 1 allows resistance only (or close to)
            # constructions to be modelled.
            coeffs = self._approximate_response_functions()
            den = np.poly1d(coeffs[-1])
            arr: List[np.ndarray] = []
            for numerator_coeff in coeffs[:-1]:
                num = np.poly1d(numerator_coeff)
                arr.extend(self._calculate_ctf(num, den, self.timestep))

            self._ctfs_internal = np.array(arr[::2] + [arr[-1]])

        return self._ctfs_internal

    def _calculate_response_matrix(self, frequencies: np.ndarray) -> np.ndarray:
        layer_matrices = [
            m.calculate_response_matrices(frequencies) for m in self.materials
        ]

        # calculate the product over all layers
        ms_matrix = reduce(np.matmul, layer_matrices)

        return ms_matrix

    def _approximate_response_functions(self) -> List[np.ndarray]:
        """Approximate the transfer functions by rational functions."""
        frequencies = _calculate_frequencies()
        response_matrix = self._calculate_response_matrix(frequencies)
        response_functions = _calculate_response_functions(response_matrix)

        # calculate the best fit for approximation
        coeff_map = defaultdict(list)
        value_map: Dict = defaultdict(float)
        for numerator_order in range(self._min_ctf_order, self._max_ctf_order + 1):
            for denominator_order in range(numerator_order, self._max_ctf_order + 1):
                key = (numerator_order, denominator_order)

                coeffs = [None] * 4

                # first calculate the cross term (Y) to fix the denominator
                try:
                    coeffs[1], coeffs[3] = _approximate_response_function(
                        response_functions[1],
                        frequencies,
                        denominator_order=denominator_order,
                        numerator_order=numerator_order,
                    )
                except np.linalg.LinAlgError:
                    # catch the case of a singular matrix solution
                    continue

                # calculate the value of the denominator at frequences
                denom_values = np.polyval(coeffs[3], frequencies)

                # we can't except blow up of the solution
                # XXX: since we need roots elsewhere it may be better to return them
                if np.any(np.real(np.roots(coeffs[3])) >= 0.0):
                    continue

                # calculate the diff with the exact
                diff = (
                    response_functions[1]
                    - np.polyval(coeffs[1], frequencies) / denom_values
                )
                value_map[key] += np.real(np.sum(diff * diff.conjugate()))

                # now calculate the other functions
                for idx, response_function in enumerate(response_functions):
                    if idx == 1:
                        continue
                    coeffs[idx], _ = _approximate_response_function(
                        response_function,
                        frequencies,
                        denominator_order=denominator_order,
                        numerator_order=numerator_order,
                        denominator_values=denom_values,
                    )

                    # add to the diff
                    diff = (
                        response_function
                        - np.polyval(coeffs[idx], frequencies) / denom_values
                    )
                    value_map[key] += np.real(np.sum(diff * diff.conjugate()))

                coeff_map[key] = coeffs

        return coeff_map[min(value_map, key=value_map.get)]

    def _calculate_ctf(
        self, numerator_poly: np.poly, denominator_poly: np.poly, timestep: int
    ) -> Tuple[np.ndarray, np.ndarray]:
        # step 4
        roots = denominator_poly.roots
        delta = _calculate_residuals(numerator_poly, denominator_poly, roots)

        # XXX: delta is complex for some constructions, not sure why
        sum_delta = np.sum(delta)
        assert np.imag(sum_delta) == 0.0
        sum_delta = np.real(sum_delta)

        coeffs = np.ones((roots.shape[0], 2), dtype=np.complex)

        # note that there is no -1.0 factor in the exponent since the paper
        # assumes the roots are -si (see eq. 35)
        coeffs[:, 0] = -1.0 * np.exp(roots * timestep)

        # XXX: this solves the issue of nans for very low resistance constructions
        coeffs[np.isinf(coeffs)] = 0.0

        fac = np.array([-1.0, 1.0])

        if coeffs.size > 0:
            # note: convolution is analogous to polynomial multiplication
            polytot = reduce(np.convolve, coeffs)
        else:
            polytot = polyone

        numerator_poly = polytot * sum_delta

        partial_numerator_sum = np.zeros_like(numerator_poly[1:], dtype=np.complex)
        for idx, delta_i in enumerate(delta):
            mask = np.ones_like(roots, dtype=bool)
            mask[idx] = False
            partial_numerator_sum += delta_i * reduce(np.convolve, coeffs[mask], 1.0)

        # XXX: delta is complex for some constructions, not sure why
        assert np.all(
            np.abs(np.imag(partial_numerator_sum[np.isfinite(partial_numerator_sum)]))
            <= _EPS
        )
        partial_numerator_sum = np.real(partial_numerator_sum)

        if partial_numerator_sum.size > 0:
            numerator_poly -= np.convolve(partial_numerator_sum, fac)

        numerator_poly = np.convolve(numerator_poly, fac)

        assert np.abs(numerator_poly[-1]) <= _EPS

        # this is a division by z
        numerator_poly = numerator_poly[:-1]

        numerator_poly += polytot * self.thermal_transmittance * timestep
        numerator_poly /= timestep

        assert np.all(
            np.abs(np.imag(numerator_poly[np.isfinite(numerator_poly)])) <= _EPS
        )
        numerator_poly = np.real(numerator_poly)

        # numerator_poly[np.abs(numerator_poly) <= _EPS] = 0.0

        assert np.all(np.abs(np.imag(polytot[np.isfinite(polytot)])) <= _EPS)
        polytot = np.real(polytot)

        return numerator_poly, polytot

    @property
    def _timeseries_length(self) -> int:
        return self._ctfs.shape[-1]  # pylint: disable=unsubscriptable-object

    def calculate_heat_flux_inside(
        self,
        outside_temps: np.ndarray,
        inside_temps: np.ndarray,
        inside_heat_fluxes: np.ndarray,
        current_index: int,
    ) -> float:
        """Calculate the inside heat flux."""
        # short circuit for resistance only constructions
        if self._is_resistance_only:
            return self.thermal_transmittance * (
                outside_temps[current_index] - inside_temps[current_index]
            )
        # determine the array values
        take = range(current_index - self._timeseries_length + 1, current_index + 1)
        outside_temps = outside_temps[take]
        inside_temps = inside_temps[take]
        inside_heat_fluxes = inside_heat_fluxes[take[:-1]]

        # from eq. 34
        qi_new = (
            np.sum(self._ctfs[1] * outside_temps)
            - np.sum(self._ctfs[3][:-1] * inside_heat_fluxes)
            - np.sum(self._ctfs[2] * inside_temps)
        )

        return qi_new

    def calculate_heat_flux_outside(
        self,
        outside_temps: np.ndarray,
        inside_temps: np.ndarray,
        outside_heat_fluxes: np.ndarray,
        current_index: int,
    ) -> float:
        """Calculate the inside heat flux."""
        # short circuit for resistance only constructions
        if self._is_resistance_only:
            return self.thermal_transmittance * (
                inside_temps[current_index] - outside_temps[current_index]
            )
        # determine the array values
        take = range(current_index - self._timeseries_length + 1, current_index + 1)
        outside_temps = outside_temps[take]
        inside_temps = inside_temps[take]
        outside_heat_fluxes = outside_heat_fluxes[take[:-1]]

        # from eq. 34
        qo_new = (
            -np.sum(self._ctfs[0] * outside_temps)
            - np.sum(self._ctfs[3][:-1] * outside_heat_fluxes)
            + np.sum(self._ctfs[2] * inside_temps)
        )

        return qo_new



import numpy as np

from ambient import air

# aim to have all calculations less than 1% error
# TODO: not currently achieving this
RTOL = 0.011

ALTITUDE = np.array(
    [
        -500,
        0,
        500,
        1000,
        1500,
        2000,
        2500,
        3000,
        4000,
        5000,
        6000,
        7000,
        8000,
        9000,
        10000,
    ]
)


def test_standard_pressure():
    """Test calculation of standard pressure from altitude."""
    pressure = 1000 * np.array(
        [
            107.478,
            101.325,
            95.461,
            89.875,
            84.556,
            79.495,
            74.682,
            70.108,
            61.640,
            54.020,
            47.181,
            41.061,
            35.600,
            30.742,
            26.436,
        ]
    )

    assert np.allclose(air.standard_pressure(ALTITUDE), pressure, rtol=0.0, atol=0.5)


def test_standard_temperature():
    """Test calculation of standard temperature from altitude."""
    temperature = 273.15 + np.array(
        [
            18.2,
            15,
            11.8,
            8.5,
            5.2,
            2,
            -1.2,
            -4.5,
            -11.0,
            -17.5,
            -24.0,
            -30.5,
            -37.0,
            -43.5,
            -50,
        ]
    )

    assert np.allclose(
        air.standard_temperature(ALTITUDE), temperature, rtol=0.0, atol=0.5
    )


def test_saturation_pressure_ice():
    temperature_ref = air.celsius_to_kelvin(np.array([-60.0, -40.0, -20.0, 0.0]))
    pressure_ref = 1000.0 * np.array([0.0010813, 0.012841, 0.10324, 0.61115])

    pressure_calc = air._saturation_pressure_ice(temperature_ref)

    assert np.allclose(pressure_ref, pressure_calc, rtol=0.001)


def test_saturation_pressure_water():
    temperature_ref = air.celsius_to_kelvin(np.array([0.0, 50.0, 100.0, 150.0]))
    pressure_ref = 1000.0 * np.array([0.61121, 12.351, 101.42, 476.10])

    pressure_calc = air._saturation_pressure_water(temperature_ref)

    assert np.allclose(pressure_ref, pressure_calc, rtol=0.001)


def test_saturated():
    """Test saturated states from RP-1485 Table A.8."""
    drybulb = air.celsius_to_kelvin(np.array([0, 20, 40, 60, 80]))

    hr = np.array([0.0037900, 0.0147605, 0.0491445, 0.1535447, 0.5529261])
    sv = np.array([0.7780, 0.8498, 0.9567, 1.1752, 1.8809])
    en = np.array([9.475, 57.558, 166.685, 460.880, 1541.765])
    humidity_ratio = np.full_like(hr, air.SATURATED_HUMIDITY_RATIO)

    ma = air.MoistAir(
        property_type=air.PropertyType.HUMIDITY_RATIO,
        property_value=humidity_ratio,
        drybulb=drybulb,
    )

    assert np.allclose(ma.humidity_ratio, hr, atol=0.0, rtol=RTOL)
    assert np.allclose(ma.specific_volume, sv, atol=0.0, rtol=RTOL)
    assert np.allclose(ma.specific_enthalpy, en, atol=0.0, rtol=RTOL)



from unittest.mock import patch

import numpy as np

from ambient.solar import SolarBase, SolarDesignDay, Location
from ambient import time

# patch ABC so we can test methods
SolarBase.__abstractmethods__=set()

# from table D.1, corrected for 1-based
DAYS_OF_YEAR = np.array((21, 52, 80, 111, 141, 172, 202, 233, 264, 294, 325, 355)) - 1

DESIGN_DAYS = time.TimePeriodBase(days=DAYS_OF_YEAR)

ATLANTA_GA = Location(latitude=33.64, longitude=84.43, timezone=-5)
SOLAR_ATLANTA_GA = SolarBase(location=ATLANTA_GA, timeperiod=DESIGN_DAYS)


def test_extraterrestrial_solar():
    """Test calculation the extraterrestrial solar."""
    paper_ets = np.array(
        (1410, 1397, 1378, 1354, 1334, 1323, 1324, 1336, 1357, 1380, 1400, 1411)
    )

    assert np.array_equal(np.round(SOLAR_ATLANTA_GA._extraterrestrial_solar), paper_ets)


def test_equation_of_time():
    """Test the calculation of the equation of time."""
    paper_eot = np.array(
        (-10.6, -14.0, -7.9, 1.2, 3.7, -1.3, -6.4, -3.6, 6.9, 15.5, 13.8, 2.2)
    )

    assert np.array_equal(np.round(SOLAR_ATLANTA_GA._equation_of_time, 1), paper_eot)


def test_declination():
    """Test the calculation of the declination angle."""
    paper_dec = np.array(
        (-20.1, -11.2, -0.4, 11.6, 20.1, 23.4, 20.4, 11.8, -0.2, -11.8, -20.4, -23.4)
    )

    assert np.array_equal(
        np.round(np.degrees(SOLAR_ATLANTA_GA._declination), 1), paper_dec
    )


def test_solar_irradiance_atlanta_georgia():
    """Test the calculation of solar irradiance."""
    loc = Location(latitude=33.64, longitude=84.43, timezone=-5)
    sol = SolarDesignDay(
        location=loc,
        timeperiod=DESIGN_DAYS,
        tau_b=np.array(
            (
                0.334,
                0.324,
                0.355,
                0.383,
                0.379,
                0.406,
                0.440,
                0.427,
                0.388,
                0.358,
                0.354,
                0.335,
            )
        ),
        tau_d=np.array(
            (
                2.614,
                2.580,
                2.474,
                2.328,
                2.324,
                2.270,
                2.202,
                2.269,
                2.428,
                2.514,
                2.523,
                2.618,
            )
        ),
    )

    paper_beam = np.array((884, 932, 922, 904, 907, 877, 846, 852, 875, 879, 847, 861))
    paper_diffuse = np.array((77, 87, 105, 127, 129, 136, 144, 133, 109, 92, 84, 73))

    assert np.allclose(
        np.round(sol._beam_normal_irradiance[:, 12]), paper_beam, rtol=0, atol=2
    )
    assert np.allclose(
        np.round(sol._diffuse_horizontal_irradiance[:, 12]),
        paper_diffuse,
        rtol=0,
        atol=1,
    )


def test_solar_irradiance_chicago_illinois():
    """Test the calculation of solar irradiance."""
    loc = Location(latitude=41.99, longitude=87.91, timezone=-6)
    sol = SolarDesignDay(
        location=loc,
        timeperiod=DESIGN_DAYS,
        tau_b=np.array(
            (
                0.288,
                0.305,
                0.325,
                0.359,
                0.369,
                0.383,
                0.415,
                0.416,
                0.392,
                0.344,
                0.306,
                0.294,
            )
        ),
        tau_d=np.array(
            (
                2.524,
                2.474,
                2.473,
                2.343,
                2.310,
                2.285,
                2.209,
                2.242,
                2.343,
                2.487,
                2.629,
                2.584,
            )
        ),
    )

    paper_beam = np.array((871, 906, 924, 909, 905, 889, 855, 843, 838, 847, 848, 836))
    paper_diffuse = np.array((70, 87, 98, 120, 127, 131, 140, 131, 111, 85, 64, 62))

    assert np.allclose(
        np.round(sol._beam_normal_irradiance[:, 12]), paper_beam, rtol=0, atol=2
    )
    assert np.allclose(
        np.round(sol._diffuse_horizontal_irradiance[:, 12]),
        paper_diffuse,
        rtol=0,
        atol=1,
    )
